/** 
 * W-DATA format
 * @author Gabriel Wlazlowski, Warsaw University of Technology, 2020
 * 
 * Tool for extracting subset from wdata set
 * */

// wdata lib
#include "wdata.h"

// other libs
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <libgen.h>
#include <math.h>
#include <complex.h>

#define cppmallocl(pointer, size, type)                                     \
    if ((pointer = (type *)malloc((size) * sizeof(type))) == NULL)          \
    {                                                                       \
        fprintf(stderr, "error: cannot malloc()! Exiting!\n");              \
        fprintf(stderr, "error: file=`%s`, line=%d\n", __FILE__, __LINE__); \
        return -1;                                                          \
    }

int main(int argc, char **argv)
{
    printf("# WDATA SUBSET EXTRACTOR\n");

    if (argc != 5 && argc != 6)
    {
        printf("# Usage: %s file.wtxt outprefix start stop\n", argv[0]);
        printf("# \tor\n");
        printf("# Usage: %s file.wtxt outprefix start stop stride\n", argv[0]);
        printf("# \tfile.wtxt    - metadata file (INPUT)\n");
        printf("# \toutprefix    - new metadata file will be written to outprefix.wtxt (OUTPUT)\n");
        printf("# \t[start,stop) - range for subtructing cycles\n");
        printf("# \tstride       - every `stride` frame will be taken only from given range, optional, default stride=1\n");
        return 0;
    }

    char ctmp1[512], ctmp2[512];
    char basedir[512];
    getcwd(basedir, 512);
    strcpy(ctmp1, argv[1]);
    strcpy(ctmp2, argv[2]);
    char *indir = dirname(ctmp1);
    char *outdir = dirname(ctmp2);

    printf("# WORKING DIR: `%s` --> `%s`\n", indir, outdir);

    int start = atoi(argv[3]);
    int stop = atoi(argv[4]);
    printf("# SUBTRUCTION RANGE: [%d,%d)\n", start, stop);

    int stride = 1;
    if (argc == 6)
        stride = atoi(argv[5]);
    printf("# STRIDE: %d\n", stride);

    // create metadata handler
    wdata_metadata md;
    wdata_metadata mdout;

    // read metadata from file
    int ierr;
    printf("# READING FILE: `%s`\n", argv[1]);
    ierr = wdata_parse_metadata_file(argv[1], &md);
    if (ierr != 0)
    {
        printf("# Cannot read metadata file!\n");
        return 1;
    }

    mdout = md;
    mdout.cycles = 0;
    mdout.t0 = md.t0 + md.dt * start;
    strcpy(mdout.prefix, basename(argv[2]));

    // set working dirs
    wdata_set_working_dir(&md, indir);
    wdata_set_working_dir(&mdout, outdir);
    wdata_clear_database(&mdout);

    double *data;
    int bdim = wdata_get_blocklength(&md); // get block size
    cppmallocl(data, bdim * 3, double);

    int icycle, ivar;
    for (icycle = 0; icycle < md.cycles; icycle++)
        if (icycle >= start && icycle < stop && icycle % stride == 0)
        {
            for (ivar = 0; ivar < md.nvar; ivar++)
            {
                ierr = wdata_read_cycle(&md, md.var[ivar].name, icycle, data);
                if (ierr != 0)
                {
                    printf("# ERROR: Cannot read datablock!\n");
                    return 1;
                }

                sprintf(mdout.var[ivar].format, "wdat"); // use wdat format
                ierr = wdata_write_cycle(&mdout, mdout.var[ivar].name, data);
                if (ierr != 0)
                {
                    printf("# ERROR: Cannot write datablock!\n");
                    return 1;
                }
            }
            wdata_add_cycle(&mdout);
        }

    // Write new metadata file
    char file_name[512];
    sprintf(file_name, "%s.wtxt", argv[2]);
    printf("# WRITING NEW METADAFILE: `%s`\n", file_name);
    FILE *f = fopen(file_name, "w");
    fprintf(f, "# Generated by: %s %s %s %s %s %d\n\n", argv[0], argv[1], argv[2], argv[3], argv[4], stride);
    wdata_print_metadata(&mdout, f);
    fclose(f);

    printf("# DONE.\n");

    return 0;
}
